from __future__ import print_function, unicode_literals

import os
import shutil
import zipfile
import datetime
import tempfile
import subprocess

import pytest
import numpy as np
from numpy.testing import assert_almost_equal
from sklearn.dummy import DummyClassifier

from destimator import DescribedEstimator, utils


@pytest.fixture
def features():
    return np.zeros([10, 3])


@pytest.fixture
def labels():
    labels = np.zeros(10)
    labels[5:] = 1.0
    return labels


@pytest.fixture
def clf(features, labels):
    clf = DummyClassifier(strategy='constant', constant=0.0)
    clf.fit(features, labels)
    return clf


@pytest.fixture
def clf_described(clf, features, labels, feature_names):
    return DescribedEstimator(clf, features, labels, features, labels, feature_names)


@pytest.fixture
def feature_names():
    return ['one', 'two', 'three']


class TestDescribedEstimator(object):
    def test_init(self, clf_described):
        assert clf_described.n_training_samples_ == 10
        assert clf_described.n_features_ == 3

    def test_init_error(self, clf, features, labels, feature_names):
        with pytest.raises(ValueError):
            wrong_labels = np.zeros([9, 1])
            DescribedEstimator(clf, features, wrong_labels, features, labels, feature_names)

        with pytest.raises(ValueError):
            wrong_feature_names = ['']
            DescribedEstimator(clf, features, labels, features, labels, wrong_feature_names)

    def test_from_file(self, clf_described):
        save_dir = tempfile.mkdtemp()
        try:
            file_path = clf_described.save(save_dir)
            destimator = DescribedEstimator.from_file(file_path)
            assert destimator == clf_described
        finally:
            shutil.rmtree(save_dir)

    def test_is_compatible(self, clf, clf_described, features, labels):
        compatible = DescribedEstimator(clf, features, labels, features, labels, ['one', 'two', 'three'])
        assert clf_described.is_compatible(compatible)

        incompatible = DescribedEstimator(clf, features, labels, features, labels, ['one', 'two', 'boom'])
        assert not clf_described.is_compatible(incompatible)

    def test_metadata(self, clf, features, labels, feature_names):
        clf_described = DescribedEstimator(clf, features, labels, features, labels, feature_names)
        d = clf_described.metadata
        assert d['feature_names'] == feature_names
        # assert type(d['metadata_version']) == str
        assert type(datetime.datetime.strptime(d['created_at'], '%Y-%m-%d-%H-%M-%S')) == datetime.datetime
        # assert type(d['vcs_hash']) == str
        assert type(d['distribution_info']) == dict
        # assert type(d['distribution_info']['python']) == str
        assert type(d['distribution_info']['packages']) == list
        assert type(d['performance_scores']['precision']) == list
        assert type(d['performance_scores']['precision'][0]) == float
        assert type(d['performance_scores']['recall']) == list
        assert type(d['performance_scores']['recall'][0]) == float
        assert type(d['performance_scores']['fscore']) == list
        assert type(d['performance_scores']['fscore'][0]) == float
        assert type(d['performance_scores']['support']) == list
        assert type(d['performance_scores']['support'][0]) == int
        assert type(d['performance_scores']['roc_auc']) == float
        assert type(d['performance_scores']['log_loss']) == float

    def test_get_metric(self, clf_described):
        assert clf_described.recall == [1.0, 0.0]
        assert clf_described.roc_auc == 0.5
        # log_loss use epsilon 1e-15, so -log(1e-15) / 2 approximately equal 20
        assert_almost_equal(clf_described.log_loss, 17.269, decimal=3)

    def test_save_classifier(self, clf_described):
        save_dir = tempfile.mkdtemp()
        try:
            saved_name = clf_described.save(save_dir)
            assert os.path.dirname(saved_name) == save_dir
            assert os.path.isfile(saved_name)
            assert saved_name.endswith('.zip')
            zf = zipfile.ZipFile(saved_name)
            files_present = zf.namelist()
            expected_files = [
                'model.bin', 'features_train.bin', 'labels_train.bin',
                'features_test.bin', 'labels_test.bin', 'metadata.json',
            ]
            # could use a set, but this way errors are easier to read
            for f in expected_files:
                assert f in files_present
        finally:
            shutil.rmtree(save_dir)

    def test_save_classifier_with_filename(self, clf_described):
        save_dir = tempfile.mkdtemp()
        try:
            saved_name = clf_described.save(save_dir, filename='boom.pkl')
            assert os.path.basename(saved_name) == 'boom.pkl.zip'
            assert os.path.isfile(saved_name)
        finally:
            shutil.rmtree(save_dir)

    def test_save_classifier_nonexistent_path(self, clf_described):
        save_dir = tempfile.mkdtemp()
        try:
            saved_name = clf_described.save(os.path.join(save_dir, 'nope'))
            os.path.dirname(saved_name) == save_dir
            assert os.path.isfile(saved_name)
        finally:
            shutil.rmtree(save_dir)


class TestGetCurrentGitHash(object):
    def test_get_current_vcs_hash(self, monkeypatch):
        def fake_check_output(*args, **kwargs):
            return b'thisisagithash'
        monkeypatch.setattr(subprocess, 'check_output', fake_check_output)
        assert utils.get_current_vcs_hash() == 'thisisagithash'

    def test_get_current_vcs_hash_no_git(self, monkeypatch):
        def fake_check_output(*args, **kwargs):
            raise OSError()
        monkeypatch.setattr(subprocess, 'check_output', fake_check_output)
        assert utils.get_current_vcs_hash() == ''

    def test_get_current_vcs_hash_git_error(self, monkeypatch):
        def fake_check_output(*args, **kwargs):
            raise subprocess.CalledProcessError(0, '', '')
        monkeypatch.setattr(subprocess, 'check_output', fake_check_output)
        assert utils.get_current_vcs_hash() == ''



from .described_estimator import DescribedEstimator


__version__ = '0.1.3'

__title__ = 'destimator'
__description__ = 'A metadata-saving proxy for scikit-learn etimators.'
__uri__ = 'https://github.com/rainforestapp/destimator'

__author__ = 'Maciej Gryka'
__email__ = 'maciej@rainforestqa.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2015 RainforestQA'



from __future__ import unicode_literals

import pip
import subprocess


def get_current_vcs_hash():
    """Return the current git hash, if it can be found, otherwise an empty string."""
    try:
        return subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('utf-8')
    except (OSError, subprocess.CalledProcessError):
        return ''


def get_installed_packages():
    return sorted(["%s==%s" % (i.key, i.version) for i in pip.get_installed_distributions()])



import codecs
import os
import re

from setuptools import setup, find_packages


###################################################################

NAME = 'destimator'
PACKAGES = find_packages(where='.')
META_PATH = os.path.join('destimator', '__init__.py')
KEYWORDS = ['scikit-learn', 'machine-learning', 'data', 'science']
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
]
INSTALL_REQUIRES = [
    'requests>=2.8.1',
    'numpy>=1.10.1',
    'scipy>=0.16.0',
    'scikit-learn>=0.17',
]

EXTRAS_REQUIRE = {
    'dev': [
        'check-manifest',
        'twine',
        'sphinx',
        'sphinx_rtd_theme',
    ],
    'test': [
        'pytest',
        'flake8',
        'pytest-cov',
        'tox',
    ],
}

###################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string.'.format(meta=meta))


if __name__ == '__main__':
    setup(
        name=NAME,
        description=find_meta('description'),
        license=find_meta('license'),
        url=find_meta('uri'),
        version=find_meta('version'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        maintainer=find_meta('author'),
        maintainer_email=find_meta('email'),
        keywords=KEYWORDS,
        long_description=read('README.rst'),
        packages=PACKAGES,
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRE,
    )



from __future__ import print_function, unicode_literals

import os
import sys
import shutil
import zipfile
import datetime
import tempfile
import json as json
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from contextlib import closing

import requests
import numpy as np
from sklearn.externals import joblib
from sklearn.metrics import precision_recall_fscore_support, roc_auc_score, log_loss

from .utils import get_current_vcs_hash, get_installed_packages


METADATA_VERSION = '2'


class DescribedEstimator(object):
    def __init__(
        self,
        clf,
        features_train, labels_train, features_test, labels_test, feature_names=None,
        *args, **kwargs
    ):
        """
        A trained classifier, together with all the data and metadata. After it
        is created, this object acts as a proxy to sklearn.base.BaseEstimator,
        meaning it forwards all calls (including predict and predict_proba) to
        the actual estimator. It also contains all the information necessary to
        recreate the classifier.

        Normally this object is initialized using:
            >> DescribedEstimator(clf, features_train, labels_train, features_test, labels_test, feature_names)

        If `compute_metadata` is set to False, a `metadata` argument must be
        given and `feature_names` is ignored.

        """
        self._clf = clf
        self._data = {
            'features_train': features_train,
            'labels_train': labels_train,
            'features_test': features_test,
            'labels_test': labels_test,
        }
        compute_metadata = 'metadata' not in kwargs
        if compute_metadata:
            precision, recall, fscore, support = precision_recall_fscore_support(
                labels_test,
                clf.predict(features_test)
            )

            roc_auc = roc_auc_score(labels_test, clf.predict_proba(features_test)[:, 1])
            self.metadata = {
                'metadata_version': METADATA_VERSION,
                'created_at': datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S'),
                'feature_names': feature_names,
                'vcs_hash': get_current_vcs_hash(),
                'distribution_info': {
                    'python': sys.version,
                    'packages': get_installed_packages(),
                },
                'performance_scores': {
                    'precision': [float(p) for p in precision],  # cast to make it
                    'recall':    [float(r) for r in recall],     # JSONserializable
                    'fscore':    [float(f) for f in fscore],
                    'support':   [int(s) for s in support],
                    'roc_auc':   float(roc_auc),
                    'log_loss':  float(log_loss(labels_test, clf.predict_proba(features_test)[:, 1])),
                }
            }
        else:
            self.metadata = kwargs.pop('metadata')
        self.validate_data()
        super(DescribedEstimator, self).__init__(*args, **kwargs)

    def __eq__(self, other):
        data_equal = all([
            np.array_equal(self._data[k], other._data[k])
            for k in self._data.keys()
        ])
        metadata_equal = all([
            self.metadata['metadata_version'] == other.metadata['metadata_version'],
            self.metadata['feature_names'] == other.metadata['feature_names'],
            self.metadata['performance_scores'] == other.metadata['performance_scores'],
        ])
        classifier_equal = (
            self._clf.__class__ == other._clf.__class__ and
            self._clf.get_params() == other._clf.get_params()
        )
        return data_equal and metadata_equal and classifier_equal

    @classmethod
    def from_file(cls, f):
        """
        Read the described classifier from file. `f` can be a path or a
        file-like object.

        """
        zf = zipfile.ZipFile(f)
        extract_dir = tempfile.mkdtemp()
        try:
            zf.extractall(extract_dir)
            data = {}
            for fn in zf.namelist():
                fn_full = os.path.join(extract_dir, fn)
                if fn == 'model.bin':
                    data['clf'] = joblib.load(fn_full)
                elif fn == 'features_train.bin':
                    data['features_train'] = joblib.load(fn_full)
                elif fn == 'labels_train.bin':
                    data['labels_train'] = joblib.load(fn_full)
                elif fn == 'features_test.bin':
                    data['features_test'] = joblib.load(fn_full)
                elif fn == 'labels_test.bin':
                    data['labels_test'] = joblib.load(fn_full)
                elif fn == 'metadata.json':
                    data['metadata'] = json.loads(open(fn_full, 'rt').read())
            return cls(
                clf=data['clf'],
                features_train=data['features_train'],
                labels_train=data['labels_train'],
                features_test=data['features_test'],
                labels_test=data['labels_test'],
                metadata=data['metadata'],
            )
        finally:
            shutil.rmtree(extract_dir)

    @classmethod
    def from_url(cls, url):
        """Read the described classifier from the given URL."""
        f = StringIO()
        f.write(requests.get(url, stream=True).content)
        return cls.from_file(f)

    def is_compatible(self, other):
        """
        Check whether this estimator is compatible with the other one (in other
        words, are their `feature_names` the same).

        """
        return self.feature_names == other.feature_names

    @property
    def n_training_samples_(self):
        return self._data['features_train'].shape[0]

    @property
    def n_features_(self):
        if hasattr(self._clf, 'n_features_'):
            return self._clf.n_features_
        else:
            return self._data['features_train'].shape[1]

    def save(self, path, filename=None):
        """Save the classifier under `path`."""
        if filename is None:
            filename = 'destimator_' + datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        filename += '.zip'
        try:
            os.makedirs(path)
        except OSError:
            pass  # the directory exists
        archive_name = os.path.join(path, filename)
        # save files to disk (unfortunately joblib requires that)
        joblib.dump(self._clf, os.path.join(path, 'model.bin'), compress=9)
        joblib.dump(self.features_train, os.path.join(path, 'features_train.bin'), compress=9)
        joblib.dump(self.labels_train, os.path.join(path, 'labels_train.bin'), compress=9)
        joblib.dump(self.features_test, os.path.join(path, 'features_test.bin'), compress=9)
        joblib.dump(self.labels_test, os.path.join(path, 'labels_test.bin'), compress=9)
        with open(os.path.join(path, 'metadata.json'), 'wt') as f:
            f.write(json.dumps(self.metadata))
        # pack all the files into a zip archive
        with closing(zipfile.ZipFile(archive_name, 'w', zipfile.ZIP_DEFLATED)) as zf:
            for fn in ['model.bin', 'features_train.bin', 'labels_train.bin', 'features_test.bin', 'labels_test.bin', 'metadata.json']:
                zf.write(os.path.join(path, fn), fn)
        return archive_name

    def validate_data(self):
        if len(self._data['features_train']) != len(self._data['labels_train']):
            raise ValueError('Number of features and labels must be the same.')

        features_mismatch = self.features_train.shape[1] != self.features_test.shape[1]
        labels_dims_train = self.labels_train.shape[1] if self.labels_train.ndim > 1 else 1
        labels_dims_test = self.labels_test.shape[1] if self.labels_test.ndim > 1 else 1
        labels_mismatch = labels_dims_train != labels_dims_test
        if features_mismatch or labels_mismatch:
            raise ValueError('Training and test data must have the same dimensionality.')

        if self.feature_names is not None and self.features_train.shape[1] != len(self.feature_names):
            raise ValueError('Number of feature_names must match the number of features.')

        if 'feature_names' not in self.metadata:
            raise ValueError('feature_names mist be provided.')

    def __getattr__(self, name):
        if name in self._data.keys():
            return self._data[name]
        elif name in self.metadata.keys():
            return self.metadata[name]
        elif name in self.metadata['performance_scores'].keys():
            return self.metadata['performance_scores'][name]
        else:
            return getattr(self._clf, name)
